{** This unit provides some classes to interface with the BasicCard }
unit ZCDCI1;
// Version 1.0
{
Copyright (C) 2002 CHZ-Soft
You have a royalty-free right to use, modify, reproduce and 
distribute the files under this license or any modified
version or any derived work in any way you find useful, 
provided that: 

a) You agree that CHZ-Soft has no warranty, obligations or 
   liability for any of these files.

b) When distributing in source form: this license notice
   remains intact.

c) When distributing in binary form: you include one of the 
   following copyright notices in the program or documentation:
   1.) Copyright (C) 2002 CHZ-Soft 
   2.) Portions Copyright (C) 2002 CHZ-Soft
}

interface

{$R *.dcr}

uses
  Windows, Messages, SysUtils, Classes, ExtCtrls, ZCBCI_PAS;

type
{** This class interfaces to Zeitcontrol's BasicCard API.
    It provides many functions from the CRI and BCI DLLs.<br>
    If not documented otherwise the functions will return True if
    they succeeded and False if they failed. The properties Error,
    SW1SW2 and FileError will be set if appropiate.
}
  TBasicCard = class(TComponent)
  private
    fOnCardInserted: TNotifyEvent;
    fOnCardRemoved: TNotifyEvent;
    fOnReaderBusy: TNotifyEvent;
    fOldState: ZCCRIRET;
    fTimer: TTimer;

    fIsOpen: Boolean;
    fConnected: Boolean;

    fATR: ZCCRIATR;

    fMySeed: LongInt;

    procedure fTimerProc(Sender: TObject);
    function GetReaderName(Num: Integer): String;
    function GetReaderCount: Integer;
    function GetDefaultReader: String;
    procedure SetDefaultReader(Value: String);

    function GetApplicationID: String;
  protected
    fError:   ZCCRIRET;
    fFileError: Byte;
    fSW1SW2:  Word;
    fReader:  ZCCRIREADER;
    fCard:    ZCBCICARD;
  public
{** The name of the default card reader. May be set to change it. }
    property DefaultReader: String read GetDefaultReader write SetDefaultReader;  
{** The number of readers known to the API }
    property ReaderCount: Integer read GetReaderCount;
{** An array containing the names of the known readers }    
    property Readers[Num: Integer]: String read GetReaderName;
{** Last error returned by an API call }
    property Error: ZCCRIRET read fError;
{** Last FileErr returned by the BasicCard }
    property FileError: Byte read fFileError;
{** Last SW1SW2 return value from the card }
    property SW1SW2: Word read fSW1SW2;
{** The handle of the open reader. You may need this to call the API directly. }
    property ReaderHandle: ZCCRIREADER read fReader;
{** The handle of the card. You may need this to call the API directly. }
    property CardHandle: ZCBCICARD read fCard;

{** Is true when a reader is open }
    property IsOpen: Boolean read fIsOpen;
{** Is true when connected to a Card }
    property IsConnected: Boolean read fConnected;
{** The ATR returned by the card }
    property ATR: ZCCRIATR read fATR;
{** The ApplicationID of the BasicCard or an empty string if not available }
    property ApplicationID: String read GetApplicationID;


    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    // Reader functions
    function CardInReader: Boolean;
    function OpenReader(Reader: String): Boolean;
    function CloseReader: Boolean;
    function SelectReaderDialog: Boolean;

    // Card functions
    function Connect: Boolean; virtual;
    function Reconnect: Boolean; virtual;
    function Disconnect: Boolean; virtual;

    // BasicCard functions
    function Echo(var Chars: String; Incr: Byte): Boolean;
    function SetKey(KeyNum: Byte; Key: String): Boolean;
    function SetPoly(PolyA, PolyB: DWORD): Boolean;
    function StartEncryption(Algo, Key: Byte): Boolean;
    function EndEncryption: Boolean;
    function Transaction(CLA, INS: Byte; PData: PZCBCIPARM; DataCount: Byte): Boolean;
    function Transaction2(CLA, INS, P1, P2, Lc, Le: Byte; PData: PZCBCIPARM; DataCount: Byte): Boolean;

    // file functions
    function OpenFile(Path: String; Mode: Byte; RecordLen: Word; var FileNum: WORD): Boolean;
    function CloseFile(FileNum: Byte): Boolean;
    function CloseAllFiles: Boolean;
    function GetFilePos(FileNum: Byte; var FilePos: DWORD): Boolean;
    function SetFilePos(FileNum: Byte; FilePos: DWORD): Boolean;
    function FileGet(FileNum: Byte; var Data; Count: Byte): Boolean;
    function FilePut(FileNum: Byte; const Data; Count: Byte): Boolean;
    function FileLength(FileNum: Byte; var Length: DWORD): Boolean;
    function QueryEOF(FileNum: Byte; var EOF: Boolean): Boolean;
    function FileRead(FileNum: Byte; var Data; Count: Byte): Boolean;
    function FileWrite(FileNum: Byte; const Data; Count: Byte): Boolean;

    function EraseFile(Path: String): Boolean;
    function Rename(OldPath, NewPath: String): Boolean;

    // directory functions
    function ChDir(Path: String): Boolean;
    function CurDir(var Path: String): Boolean;
    function MkDir(Path: String): Boolean;
    function RmDir(Path: String): Boolean;

    // attribute and locking functions
    function GetAttr(Path: String; var Attrib: Word): Boolean;
    function GetLockInfo(Path: String; var LockInfo: ZCBCILOCKINFO): Boolean;
    function SetFileLock(Path: String; LockType: Byte; Lock: Boolean; KeyCount, Key1, Key2: Byte): Boolean; 

    function DirCount(FileSpec: String; var Count: Word): Boolean;
    function DirFile(FileSpec: String; Num: Word; var FileName: String): Boolean;
  published
{** This event is called when a card is inserted into the currently open reader }
    property OnCardInserted: TNotifyEvent read fOnCardInserted write fOnCardInserted;
{** This event is called when the card is removed from the open reader }
    property OnCardRemoved:  TNotifyEvent read fOnCardRemoved write fOnCardRemoved;
{** This event is called when the reader becomes busy, i.e. is used by another application }
    property OnReaderBusy:   TNotifyEvent read fOnReaderBusy write fOnReaderBusy;
  end;

{** TCardStream is a TStream descendant that can be used to access
    (binary) files on the BasicCard.<br>
    It will raise an ECardStreamError if something goes wrong.
    @see TStream documentation in Delphi help files
}
  TCardStream = class(TStream)
  protected
    fFileNum: Word;
    fCard: TBasicCard;
  public
    property FileNum: Word read fFileNum;
    constructor Create(Card: TBasicCard; FileName: String; Mode: Byte);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function EOF: Boolean;
  end;

{** ECardStreamError is raised when an error occurs in TCardStream.
}
  ECardStreamError = class(Exception)
  end;

procedure Register;

implementation

procedure TBasicCard.fTimerProc(Sender: TObject);
begin
 if not IsOpen then Exit;
 fTimer.Enabled:= False;
 CardInReader;
 if fError<>fOldState then begin
  fOldState:= fError;
  case fError of
   ZCCRI_NOERROR: if Assigned(fOnCardInserted) then fOnCardInserted(Self);
   ZCCRI_ERROR_NO_CARD: begin
    if Assigned(fOnCardRemoved) then fOnCardRemoved(Self);
    Disconnect;          // will fail but is needed
    fConnected:= False;  // no card = no connection
   end;
   ZCCRI_ERROR_READER_BUSY: if Assigned(fOnReaderBusy) then fOnReaderBusy(Self);
  end;
 end;
 fTimer.Enabled:= True;
end;
  
function TBasicCard.GetReaderName(Num: Integer): String;
begin
 SetLength(Result, ZCCRI_MAX_READER_NAME);
 if ZCCriGetReaderName(Num, PChar(Result))<>ZCCRI_NOERROR then
  Result:= '';
 Result:= String(PChar(Result)); // shorten it 
end;

function TBasicCard.GetReaderCount: Integer;
begin
 ZCCriGetReaderCount(@Result);
end;

function TBasicCard.GetDefaultReader: String;
begin
 SetLength(Result, ZCCRI_MAX_READER_NAME);
 if ZCCriGetDefaultReader(PChar(Result))<>ZCCRI_NOERROR then
  Result:= '';
 Result:= String(PChar(Result));  
end;

procedure TBasicCard.SetDefaultReader(Value: String);
begin
 ZCCRISetDefaultReader(PChar(Value));
end;

constructor TBasicCard.Create(AOwner: TComponent);
var YourSeed: Longint;
begin
 inherited Create(AOwner);
 fReader:= 0;
 fCard:= 0;
 fOldState:= ZCCRI_ERROR_NO_CARD;
 fOnCardInserted:= nil;
 fOnCardRemoved:= nil;
 fOnReaderBusy:= nil;
 fTimer:= TTimer.Create(Self);
 fTimer.Interval:= 250;
 fTimer.OnTimer:= fTimerProc;
 YourSeed:= RandSeed; // don't change program's randseed
 Randomize;
 fMySeed:= RandSeed;
 RandSeed:= YourSeed;
end;

destructor TBasicCard.Destroy;
begin
 if IsConnected then Disconnect;
 if IsOpen then CloseReader;
 fTimer.Free;
 inherited Destroy;
end;

{** Returns true when there is a card in the (open) reader }
function TBasicCard.CardInReader: Boolean;
begin
 fError:= ZCCriWaitCard(fReader, 0);
 Result:= fError=ZCCRI_NOERROR;
end;

{** Opens a reader and attaches to it to the BCI.
    @param Reader The name of the reader or an empty string to open the default reader }
function TBasicCard.OpenReader(Reader: String): Boolean;
begin
 Result:= False;
 if IsOpen then Exit;
 ZCCriSetOptions2(nil, ZCCRI_OPTION_FASTINIT);
 fError:= ZCCriOpenReader(@fReader, PChar(Reader));
 if fError=ZCCRI_NOERROR then
  fError:= ZCBCIAttach(fReader, @fCard);
 Result:= fError=ZCCRI_NOERROR;
 if Result then
  fTimer.Enabled:= True;
 fIsOpen:= Result;
end;

{** Detaches and closes the reader }
function TBasicCard.CloseReader: Boolean;
begin
 fError:= ZCBciDetach(fCard);
 if fError=ZCCRI_NOERROR then
  fCard:= 0;
 fError:= ZCCriCloseReader(fReader);
 Result:= fError=ZCCRI_NOERROR;
 if Result then begin
  fReader:= 0;
  fTimer.Enabled:= False;
  fIsOpen:= False;
 end;
end;

{** Opens a dialog where the user can select a reader and sets new reader if appropiate }
function TBasicCard.SelectReaderDialog: Boolean;
var Temp: String;
begin
 SetLength(Temp, ZCCRI_MAX_READER_NAME);
 fError:= ZCCriSelectReaderDialog(PChar(Temp));
 if fError=ZCCRI_NOERROR then
  fError:= ZCCRISetDefaultReader(PChar(Temp));
 Result:= fError=ZCCRI_NOERROR;
end;

{** Connects to the card }
function TBasicCard.Connect: Boolean;
begin
 Result:= False;
 if fConnected then Exit;
 fError:= ZCCriConnect(fReader, @fATR);
 Result:= fError = ZCCRI_NOERROR;
 fConnected:= Result;
end;

{** Reconnects to the card, i.e. resets it }
function TBasicCard.Reconnect: Boolean;
begin
 fError:= ZCCriReconnect(fReader, @fATR);
 Result:= fError = ZCCRI_NOERROR;
 fConnected:= Result;
end;

{** Disconnects from the card }
function TBasicCard.Disconnect: Boolean;
begin
 EndEncryption; // can't hurt
 fError:= ZCCriDisconnect(fReader);
 Result:= fError = ZCCRI_NOERROR;
 fConnected:= not Result;
end;

function TBasicCard.GetApplicationID: String;
begin
 SetLength(Result, 255);
 if ZCBciGetApplicationID(fCard, @fSW1SW2, PChar(Result), Length(Result))<>ZCBCI_NOERROR then
  Result:= '';
 Result:= String(PChar(Result)); // shorten it 
end;

{** Executes the built-in Echo function
    @param Chars The String to pass to the function and the result
    @param Incr  Increment all chars by this value }
function TBasicCard.Echo(var Chars: String; Incr: Byte): Boolean;
var Temp: String;
begin
 Temp:= Copy(Chars, 1, Length(Chars)); // make a copy
 fError:= ZcBciEcho(fCard, @fSW1SW2, Incr, PChar(Temp));
 Chars:= Copy(Temp, 1, Length(Temp));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Sets an encryption key
    @param KeyNum Key number
    @param Key    The key itself (8 or 16 bytes long) }
function TBasicCard.SetKey(KeyNum: Byte; Key: String): Boolean;
begin
 fError:= ZCBciSetKey(fCard, KeyNum, PChar(Key), Length(Key));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Sets the encryption polynoms
    @param PolyA First polynom
    @param PolyB Second polynom }
function TBasicCard.SetPoly(PolyA, PolyB: DWORD): Boolean;
begin
 fError:= ZCBciSetPoly(fCard, PolyA, PolyB);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Starts encrypted communication with card
    @param Algo The algorithm or 0 for auto-selection
    @param Key The key number }
function TBasicCard.StartEncryption(Algo, Key: Byte): Boolean;
var YourSeed: Longint;
begin
 YourSeed:= RandSeed; // use our own seed
 RandSeed:= fMySeed;
 fError:= ZCBciStartEncryption(fCard, @fSW1SW2, Algo, Key, Random(1 shl 15));
 fMySeed:= RandSeed;
 RandSeed:= YourSeed;
 Result:= fError = ZCCRI_NOERROR;
end;

{** Terminates encrypted communication }
function TBasicCard.EndEncryption: Boolean;
begin
 fError:= ZCBciEndEncryption(fCard, @fSW1SW2);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Calls a card command
    @param CLA Command class byte
    @param INS Command instruction byte
    @param PData A pointer to an array of ZBCIPARM
    @param DataCount Numer of parameters to pass
    @see API documentation for structure of ZBCIPARM }
function TBasicCard.Transaction(CLA, INS: Byte; PData: PZCBCIPARM; DataCount: Byte): Boolean;
begin
 fError:= ZCBciTransaction(fCard, CLA, INS, PData, DataCount, @fSW1SW2);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Calls a card command
    @param CLA Command class byte
    @param INS Command instruction byte
    @param P1 P1 byte
    @param P2 P2 byte
    @param Lc Lc or ZBCI_DEFAULT_LC
    @param Le Le, ZCBCI_DISABLE_LE or ZCBCI_DEFAULT_LE
    @param PData A pointer to an array of ZBCIPARM
    @param DataCount Numer of parameters to pass
    @see API documentation for structure of ZBCIPARM }
function TBasicCard.Transaction2(CLA, INS, P1, P2, Lc, Le: Byte; PData: PZCBCIPARM; DataCount: Byte): Boolean;
begin
 fError:= ZCBciTransaction2(fCard, CLA, INS, P1, P2, Lc, Le, PData, DataCount, @fSW1SW2);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Opens a file on the BasicCard
    @param Path The file name (without @:)
    @param Mode File mode
    @param RecordLen Record/block size
    @param FileNum Will contains the file number when returning successfully
    @see API documentation for valid file modes }
function TBasicCard.OpenFile(Path: String; Mode: Byte; RecordLen: Word; var FileNum: WORD): Boolean;
begin
 fError:= ZCBciOpenFile(fCard, @fSW1SW2, @fFileError, PChar(Path), Mode, RecordLen, @FileNum);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Closes a file }
function TBasicCard.CloseFile(FileNum: Byte): Boolean;
begin
 fError:= ZCBciCloseFile(fCard, @fSW1SW2, @fFileError, FileNum);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Closes all open files }
function TBasicCard.CloseAllFiles: Boolean;
begin
 fError:= ZCBciCloseAllFiles(fCard, @fSW1SW2, @fFileError);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Returns the current file position
    @param FilePos Returns current position (starting with 1) }
function TBasicCard.GetFilePos(FileNum: Byte; var FilePos: DWORD): Boolean;
begin
 fError:= ZCBciGetFilepos(fCard, @fSW1SW2, @fFileError, FileNum, @FilePos);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Sets the current file position
    @param FilePos New position (starting with 1) }
function TBasicCard.SetFilePos(FileNum: Byte; FilePos: DWORD): Boolean;
begin
 fError:= ZCBciSetFilepos(fCard, @fSW1SW2, @fFileError, FileNum, FilePos);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Reads from a binary file
    @param Data A buffer for the data
    @param Count Number of bytes to read }
function TBasicCard.FileGet(FileNum: Byte; var Data; Count: Byte): Boolean;
begin
 fError:= ZCBciFileGet(fCard, @fSW1SW2, @fFileError, FileNum, PChar(@Data), Count);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Writes to a binary file
    @param Data A buffer for the data
    @param Count Number of bytes to write }
function TBasicCard.FilePut(FileNum: Byte; const Data; Count: Byte): Boolean;
begin
 fError:= ZCBciFilePut(fCard, @fSW1SW2, @fFileError, FileNum, PChar(@Data), Count);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Returns the file length
    @param Length Will contain the file length }
function TBasicCard.FileLength(FileNum: Byte; var Length: DWORD): Boolean;
begin
 fError:= ZCBciFileLength(fCard, @fSW1SW2, @fFileError, FileNum, @Length);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Checks if end of file has been reached
    @param EOF Will contain true if EOF reached when returning successfully }
function TBasicCard.QueryEOF(FileNum: Byte; var EOF: Boolean): Boolean;
begin
 fError:= ZCBciQueryEOF(fCard, @fSW1SW2, @fFileError, FileNum, @EOF);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Reads from a non-binary file
    @param Data A buffer for the data
    @param Count Number of bytes to read }
function TBasicCard.FileRead(FileNum: Byte; var Data; Count: Byte): Boolean;
begin
 fError:= ZCBciFileRead(fCard, @fSW1SW2, @fFileError, FileNum, PChar(@Data), Count);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Writes to a non-binary file
    @param Data A buffer for the data
    @param Count Number of bytes to write }
function TBasicCard.FileWrite(FileNum: Byte; const Data; Count: Byte): Boolean;
begin
 fError:= ZCBciFileWrite(fCard, @fSW1SW2, @fFileError, FileNum, PChar(@Data), Count);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Erases a card file
    @param Path File name }
function TBasicCard.EraseFile(Path: String): Boolean;
begin
 fError:= ZCBciEraseFile(fCard, @fSW1SW2, @fFileError, PChar(Path));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Renames or moves a file or directory
    @param OldPath Current filename
    @param NewPath New filename }
function TBasicCard.Rename(OldPath, NewPath: String): Boolean;
begin
 fError:= ZCBciRename(fCard, @fSW1SW2, @fFileError, PChar(OldPath), PChar(NewPath));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Changes current directory on card
    @param Path New directory }
function TBasicCard.ChDir(Path: String): Boolean;
begin
 fError:= ZCBciChDir(fCard, @fSW1SW2, @fFileError, PChar(Path));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Returns current directory on card
    @param Path Will contain current directory }
function TBasicCard.CurDir(var Path: String): Boolean;
begin
 SetLength(Path, 255);
 fError:= ZCBciCurDir(fCard, @fSW1SW2, @fFileError, PChar(Path), 255);
 Path:= String(PChar(Path));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Creates a new directory on card
    @param Path New directory }
function TBasicCard.MkDir(Path: String): Boolean;
begin
 fError:= ZCBciMkDir(fCard, @fSW1SW2, @fFileError, PChar(Path));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Removes a directory from card
    @param Path Directory to remove }
function TBasicCard.RmDir(Path: String): Boolean;
begin
 fError:= ZCBciRmDir(fCard, @fSW1SW2, @fFileError, PChar(Path));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Counts files and directories matching specification
    @param FileSpec File specification
    @param Count Will contain number of files and directories }
function TBasicCard.DirCount(FileSpec: String; var Count: Word): Boolean;
begin
 fError:= ZCBciDirCount(fCard, @fSW1SW2, @fFileError, PChar(FileSpec), @Count);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Returns file or directory name of a specific file
    @param FileSpec File specification
    @param Num Number of file (starting with 1)
    @param FileName Returns with name of matched file
    @see DirCount to get number of files }
function TBasicCard.DirFile(FileSpec: String; Num: Word; var FileName: String): Boolean;
begin
 SetLength(FileName, 255);
 fError:= ZCBciDirFile(fCard, @fSW1SW2, @fFileError, PChar(FileSpec), Num, PChar(FileName), 255);
 FileName:= String(PChar(FileName));
 Result:= fError = ZCCRI_NOERROR;
end;

{** Gets file or directory attributes
    @param Path File name
    @param Attrib If successful will return attribute }
function TBasicCard.GetAttr(Path: String; var Attrib: Word): Boolean;
begin
 fError:= ZCBciGetAttr(fCard, @fSW1SW2, @fFileError, PChar(Path), @Attrib);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Gets lock information of file or directory
    @param Path File name
    @param LockInfo ZCBCILOCKINFO structure that will contain lock information
    @see API documentation for ZCBCILOCKINFO structure }
function TBasicCard.GetLockInfo(Path: String; var LockInfo: ZCBCILOCKINFO): Boolean;
begin
 fError:= ZCBciGetLockInfo(fCard, @fSW1SW2, @fFileError, PChar(Path), @LockInfo);
 Result:= fError = ZCCRI_NOERROR;
end;

{** Sets lock information of file or directory
    @param Path File name
    @param LockType Lock type (ZCBCI_LOCK_...)
    @param Lock True to lock file, False to unlock it
    @param KeyCount Number of access keys (0..2)
    @param Key1 Number of first key (if appropiate)
    @param Key2 Number of second key (if appropiate)
    @see API documentation }
function TBasicCard.SetFileLock(Path: String; LockType: Byte; Lock: Boolean; KeyCount, Key1, Key2: Byte): Boolean;
begin
 fError:= ZCBciSetFileLock(fCard, @fSW1SW2, @fFileError, PChar(Path), LockType, Lock, KeyCount, Key1, Key2);
 Result:= fError = ZCCRI_NOERROR;
end;

// -------  TCardStream  -------

{** Opens a file on the BasicCard
    @param Card The initialised and connected BasicCard object
    @param FileName The name of the file
    @param Mode The file mode (just read and/or write, access is always set to binary }
constructor TCardStream.Create(Card: TBasicCard; FileName: String; Mode: Byte);
begin
 inherited Create;
 fCard:= Card;
 if not fCard.OpenFile(FileName, (Mode And $f) Or ZCBCI_MODE_BINARY, 1, fFileNum) then
  raise ECardStreamError.Create('OpenFile('+FileName+')');
end;

{** Closes the file }
destructor TCardStream.Destroy;
begin
 fCard.CloseFile(fFileNum);
 inherited Destroy;
end;

function Min(a,b: DWORD): DWORD;
begin
 if a<b then Result:= a else Result:= b;
end;

{** Reads from file
    @see TStream.Read documentation }
function TCardStream.Read(var Buffer; Count: Longint): Longint;
var FL, FP, RB: DWORD; TR: Longint; P: Pointer;
begin
 fCard.GetFilePos(fFileNum, FP);
 fCard.FileLength(fFileNum, FL);
 RB:= FL-FP+1; // available bytes for read
 Result:= Min(Count, RB);
 TR:= Result;
 P:= @Buffer;
 repeat // read in chunks of 120 bytes
  if not fCard.FileGet(fFileNum, P^, min(TR,120)) then
   raise ECardStreamError.Create('Read');
  P:= Pointer(LongWord(P) + 120);
  TR:= TR - 120;
 until TR<=0;
end;

{** Sets new file position
    @see TStream.Seek documentation }
function TCardStream.Seek(Offset: Longint; Origin: Word): Longint;
var D: DWORD;
begin
 case Origin of
  soFromBeginning: fCard.SetFilePos(fFileNum, Offset+1);
  soFromCurrent: begin
                  fCard.GetFilePos(fFileNum, D);
                  fCard.SetFilePos(fFileNum, D+Offset);
                 end;
  soFromEnd    : begin
                  fCard.FileLength(fFileNum, D);
                  fCard.SetFilePos(fFileNum, D+Offset+1);
                 end;
 end;
 fCard.GetFilePos(fFileNum, D);
 Result:= D-1;
end;

{** Writes to file
    @see TStream.Write documentation }
function TCardStream.Write(const Buffer; Count: Longint): Longint;
var TW: Longint; P: Pointer;
begin
 Result:= Count;
 TW:= Result;
 P:= @Buffer;
 repeat
  if not fCard.FilePut(fFileNum, P^, min(TW,120)) then
   raise ECardStreamError.Create('Write');
  P:= Pointer(LongWord(P) + 120);
  TW:= TW - 120;
 until TW<=0;
end;

{** Returns true if EOF has been reached }
function TCardStream.EOF: Boolean;
begin
 fCard.QueryEOF(fFileNum, Result);
end;

procedure Register;
begin
  RegisterComponents('BasicCard', [TBasicCard]);
end;

end.
